/**
 * bginfo library
 *
 * See brief document and version history in bginfo.h
 *
 * This is a GPL licensed software.  See the file COPYING for complete
 * details on your rights to copy, modify, and use this software.
 *
 * Copyright: Argonne National Laboratory, Department of Energy, and
 * University of Chicago.  2004, 2005 and 2006
 *
 * Written by Kazutomo Yoshii <kazutomo@mcs.anl.gov>
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#include "mpi.h"

#include "bginfo.h"


static bginfo_struct bginfo;
static int  ion_vec_capacity = 1024;

/*======================================================================
  Version macro and functions
  ======================================================================*/

int bginfo_driver_version(char* buf, int len)
{
  int   ret;
  int   found;
  char  linebuf[160];
  FILE* fp;
  char* p;
  const char* rcdist_fn = "/etc/rc.dist";

  /* get BGL_DISTDIR value */
  fp = fopen( rcdist_fn, "r" );
  if( !fp ) {
    printf("Unable to open %s\n",rcdist_fn);
    return -1;
  }
  found = 0;
  while( fgets(linebuf,sizeof(linebuf), fp)!=(char*)0 ) {
    /* eliminate comment */
    p=strchr( linebuf, '#' );
    if( p ) *p = '\0';
    if( strstr( linebuf, "BGL_DISTDIR=" ) ) {
      found ++;
      break;
    }
  }
  fclose(fp);
  if( ! found ) {
    printf("Unable to find BGL_DISTDIR value\n");
    return -1;
  }

  if( strstr( linebuf, "/ppcfloor/") ) {
    /* get driver version from link target */
    char tmpbuf[80];
    if( readlink("/bgl/BlueLight/ppcfloor", tmpbuf, sizeof(tmpbuf) ) > 0 ) {
      /*
       * there are two patterns
       * /bgl/BlueLight/ppcfloor -> V1R2M1_020_2006-060110/ppc
       * /bgl/BlueLight/ppcfloor -> /bgl/BlueLight/V1R2M1_020_2006-060110/ppc
       */

      p = strstr(tmpbuf, "/ppc");
      if( p ) {
	char* prefix = "/bgl/BlueLight/";
	*p = '\0';
	if( strncmp(tmpbuf,prefix,strlen(prefix))==0 ) {
	  strncpy( buf, tmpbuf+strlen(prefix), len );
	  buf[len-1] = 0;
	} else {
	  strncpy( buf, tmpbuf, len );
	  buf[len-1] = 0;
	}

	return strlen(buf);
      } else {
	printf("Unable to find driver\n");
	return -1;
      }
    } else {
      printf("Unable to readlink /bgl/BlueLight/ppcfloor\n");
      return -1;
    }
  } else {
    /* get driver version from BGL_DISTDIR value
     * i.e. BGL_DISTDIR="/bgl/BlueLight/V1R2M0_460_2005-051117/dist"
     */
    p = strstr( linebuf, "/bgl/BlueLight/" );
    if( !p ) {
      printf("BGL_DISTDIR value is not correct\n");
      return -1;
    }

    p += strlen("/bgl/BlueLight/");
    strncpy( buf, p, len );
    buf[len-1]=0;
    p = strstr(buf, "/dist");
    if( !p ) {
      printf("BGL_DISTDIR value is not correct\n");
      return -1;
    }
    *p = '\0';
    return strlen(buf);
  }

  return ret;
}



int bginfo_zepto_version(char* buf, int len)
{
  int rc;
  char tmpbuf[80];

  rc = bginfo_get_valstr_from_file("/etc/ZeptoOS-config.params",
				   "ZEPTO_VERSION",
				   tmpbuf, sizeof(tmpbuf) );
  if( rc < 0 ) return -1;

  if( tmpbuf[0]=='"' ) {
    char* p;
    strncpy( buf, tmpbuf+1, len);
    buf[len-1] = 0;
    p = buf;
    while(*p++) { if(*p=='"') { *p=0; break; } }
  } else {
    strncpy( buf, tmpbuf, len);
    buf[len-1] = 0;
  }

  return strlen(buf);
}


/*======================================================================
  personality related functions
  ======================================================================*/


int bginfo_is_virtualnodemode()
{
  return BGLPersonality_virtualNodeMode(&bginfo.p);
}

const char* bginfo_cn_mode()
{
  if( bginfo_is_virtualnodemode() ) return "virtual";
  return "coprocessor";
}


int bginfo_num_computenodes()
{
  return BGLPersonality_numComputeNodes(&bginfo.p);
}

int bginfo_num_ionodes()
{
  return BGLPersonality_numIONodes(&bginfo.p);
}


size_t bginfo_memorysize_per_node()
{
  size_t s;
  s = (size_t)BGLPersonality_DDRSize(&bginfo.p);
  if( bginfo_is_virtualnodemode() ) return s/2;
  return s;
}


int bginfo_get_valstr_from_file(const char* fn,
				const char* key,
				char* valstr, size_t valstrsize )
{
  FILE*  fp;
  char   buf[80];
  int    key_len = strlen(key);
  int    found;

  fp = fopen(fn, "r");
  if( !fp ) {
    bginfo_perror("fopen");
    return -1;
  }
  found = 0;
  while( fgets(buf,sizeof(buf),fp) != NULL ) {
    if( strlen(buf)>0 )    buf[strlen(buf)-1] = 0;     /* chomp */

    if( strlen(buf) > key_len+2 && 
	buf[key_len] == '=' &&
	strncmp(key, buf, key_len) == 0 ) {
      strncpy(valstr, buf+key_len+1, valstrsize );
      valstr[valstrsize-1] = 0;
      found = 1;
    }
  }
  fclose(fp);
  if( !found ) {
    printf("Unable to find %s's value\n", key);
    return -1;
  }

  return strlen(valstr);
}


int bginfo_personality_get_valstr(const char* key,
				  char* valstr, size_t valstrsize )
{
  return bginfo_get_valstr_from_file( "/proc/personality.sh",
				      key,
				      valstr, valstrsize );
}



int bginfo_ion_proc_readline(const char* fn, char* buf, int len)
{
  int fd;

  bzero( buf, len );

  fd = open(fn, O_RDONLY);
  if( fd < 0 ) {
    bginfo_perror("open");
    return -1;
  }
  read(fd, buf, len);
  close(fd);

  if( strlen(buf) > 0 ) buf[strlen(buf)-1] = 0;

  return strlen(buf);
}


/*======================================================================
  ION CN mapping functions
  ======================================================================*/

/**
 * This function re-initialize the bginfo object and the heart of
 * ION-CN mapping function family. It's mostly internal use.
 *
 * There is also bginfo_cond_create_ci_map() that only run if not
 * initialized.
 *
 * NOTE: this function must be called from all nodes because it uses
 * MPI_Allgather(). Otherwise, the parallel process stops.
 *
 * On sucess, return zero. On error, return non-zero.
 *
 */
static int bginfo_create_ci_map();       

static int bginfo_cond_create_ci_map()  /* also mostly internal use */ 
{
  if( ! bginfo.is_map_created ) {
    if( bginfo_create_ci_map() == 0 ) {
      bginfo.is_map_created = 1;
      return 0;
    } else 
      return -1;
  }
  return 0;
}



int bginfo_init()
{
  if( bginfo_cond_create_ci_map() < 0 ) return -1;

  rts_get_personality(&bginfo.p, sizeof(bginfo.p));

  return 0;
}


bginfo_struct* bginfo_get_struct()
{
  if( !bginfo.is_map_created ) return NULL;
  return &bginfo;
}


int bginfo_uneven_ci_ratio()
{
  if( !bginfo.is_map_created ) return -1;

  if( bginfo.n_ions > 1 && bginfo.ratio_uneven ) return 1;

  return 0;
}

int bginfo_ci_ratio()
{
  if( !bginfo.is_map_created ) return -1;

  return bginfo.ci_ratio;
}

int bginfo_n_ions()
{
  if( !bginfo.is_map_created ) return -1;

  return bginfo.n_ions;
}


char* bginfo_ion_ipstr( int ion_idx )
{
  static char localbuffer[16];

  if( !bginfo.is_map_created ) return NULL;

  if( ion_idx<0 || ion_idx>=bginfo.n_ions ) {
    return NULL;
  }

  strncpy(localbuffer, inet_ntoa(bginfo.ions[ion_idx].addr), 16 );
  localbuffer[15] = 0;

  return localbuffer;
}


int bginfo_find_ion(int rank)
{
  int i, pos, ion_idx;

  if( !bginfo.is_map_created ) return -1;

  pos = -1;
  for(i=0; i<bginfo.rank_ip_map_size; i++ ) {
    if( bginfo.rank_ip_map[i].rank == rank  ) {
      pos = i;
      break;
    }
  }
  if( pos<0 ) {
    return -1;
  }

  ion_idx = -1;
  for(i=0; i<bginfo.n_ions; i++) {
    if( pos>=bginfo.ions[i].sp && 
	pos<=bginfo.ions[i].ep ) {
      ion_idx = i;
      break;
    }
  }

  return ion_idx;
}


const char* bginfo_bglmpi_mapping()
{
  const char* key="BGLMPI_MAPPING";
  char* val;

  if(! bginfo_is_virtualnodemode() ) return 0; 

  val =  getenv(key);
  if( !val ) return "XYTZ";

  return val;
}



/**
 * this is a local function invoked from qsort() in create_ci_map()
 */
static int comp_by_ip(const void* a, const void* b)
{
  const bginfo_rank_ip_pair* pa = a;
  const bginfo_rank_ip_pair* pb = b;
  unsigned int  ia = *((unsigned int*)&pa->addr);
  unsigned int  ib = *((unsigned int*)&pb->addr);
  return ib - ia;
}



/**
 *
 */
static int bginfo_create_ci_map()
{
  bginfo_rank_ip_pair  p;
  char buf[80];
  int  size,rank;
  int  i;

  bzero(&bginfo, sizeof(bginfo));
  /* .is_map_created is also zero'ed */

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  bginfo.rank_ip_map_size = size;
  p.rank = rank;
  if( bginfo_personality_get_valstr("BGL_IP", buf, sizeof(buf)) < 0 ) {
    bginfo_perror("get_personality_get_valstr");
    return -1;
  }
  inet_aton( buf, &p.addr );


  bginfo.rank_ip_map = 
    (bginfo_rank_ip_pair *)malloc(sizeof(bginfo_rank_ip_pair)*size );

  if( !bginfo.rank_ip_map ) {
    bginfo_perror("malloc");
    return -1;
  }
  /* send the current proc's map to all nodes so that every node can
     have CN-ION map.*/
  {
    unsigned char* send_ptr;
    unsigned char* recv_ptr;
    size_t   count;

    count=sizeof(bginfo_rank_ip_pair);
    bginfo.rank_ip_map[rank] = p;
    send_ptr = (void*)&bginfo.rank_ip_map[rank];
    recv_ptr = (void*)bginfo.rank_ip_map;

    MPI_Allgather( send_ptr, count, MPI_BYTE,
                   recv_ptr, count, MPI_BYTE, MPI_COMM_WORLD );
  }
  /* sort by ip address */
  qsort( bginfo.rank_ip_map, size, sizeof(bginfo_rank_ip_pair), comp_by_ip );

  bginfo.ions = 
    (bginfo_ion_vec*)malloc( sizeof(bginfo_ion_vec)*ion_vec_capacity );
  if( !bginfo.ions ) {
    bginfo_perror("malloc");
    return -1;
  }

  bginfo.n_ions = 0;   /* use it as the current position */
  for(i=0; i<size; i++ ) {
    if( i==0 ) {  /* at the top */
      bginfo.ions[bginfo.n_ions].addr = bginfo.rank_ip_map[i].addr;
      bginfo.ions[bginfo.n_ions].sp = 0;
    } else if(i==size-1) {  /* at the end */
      bginfo.ions[bginfo.n_ions].ep = i;
      bginfo.n_ions ++;
    } else if( *(unsigned int*)&bginfo.rank_ip_map[i].addr != 
	       *(unsigned int*)&bginfo.rank_ip_map[i-1].addr ) {
      /* it should be 0<i<size-1. so i-1 is guaranteed */
      /* sp is filled at the top at least */
      bginfo.ions[bginfo.n_ions].ep = i-1;
      bginfo.n_ions ++;
      bginfo.ions[bginfo.n_ions].addr = bginfo.rank_ip_map[i].addr;
      bginfo.ions[bginfo.n_ions].sp = i;
    }

    if( bginfo.n_ions+1 >= ion_vec_capacity ) {
      ion_vec_capacity *= 2;
      bginfo.ions = 
	(bginfo_ion_vec*)realloc( bginfo.ions, 
				  sizeof(bginfo_ion_vec)*ion_vec_capacity );
      if( !bginfo.ions ) {
	bginfo_perror("malloc");
	return -1;
      }
    }
  }

  // if( rank==0 ) printf("n_ions=%d\n", bginfo.n_ions); 

  /* find out the maximum ratio */
  bginfo.ci_ratio = 0;
  for(i=0; i<bginfo.n_ions; i++ ) {
    if( bginfo.ions[i].ep - bginfo.ions[i].sp+1 > bginfo.ci_ratio ) {
      bginfo.ci_ratio = bginfo.ions[i].ep - bginfo.ions[i].sp+1;
    }
  }

  /* check evenness */
  bginfo.ratio_uneven = 0;
  for(i=0; i<bginfo.n_ions; i++ ) {
    if( (bginfo.ions[i].ep - bginfo.ions[i].sp+1) != bginfo.ci_ratio ) {
      bginfo.ratio_uneven = 1;
      break;
    }
  }

#if 0
  /* log file */
  {
    bginfo_find_ion(rank)
  }
#endif
  return 0;
}


/*======================================================================
  Convenient functions
  ======================================================================*/

void _bginfo_perror(const char* msg, int lno, char* fn)
{
  perror(msg);     // this msg will be written to the .error file. 
  printf("error: %s\n", msg);   // this one will be written to the .output file
}

/**
 * 
 */
int bginfo_currentdate(char* buf,int len)
{
  time_t t;

  t = time(NULL);
  strncpy( buf, ctime(&t), len);
  buf[len-1] = 0;

  if( strlen(buf)>1 && buf[strlen(buf)-1]=='\n') {
    buf[strlen(buf)-1]=0;
  }
  return strlen(buf);
}



/**
 * Prints BG/L information (mostly I/O node related)
 * 
 * @param  level    0: no CN mapping, no ION IP, NO kernel version
 *                  1: all
 *
 * NOTE: this functions involes bginfo_init(), so you don't need to
 * call bginfo_init() by yourself.
 */
static void _bginfo_print_info(int level)
{
  int i,j;
  int size,rank;
  bginfo_struct* bgi;
  char tmpbuf[256];


  /* bginfo_init() must be called from all procs */
  if( bginfo_init()==-1 ) {
    printf("Unable to init bginfo\n");
    return;
  }

  MPI_Comm_size(MPI_COMM_WORLD, &size);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  if( rank>0 ) return;

  puts("#");

  printf("# [BG/L Information - %s ]\n", (level==0)?"basic":"full" );


  puts("#");
  if( bginfo_currentdate(tmpbuf,sizeof(tmpbuf)) > 0 ) {
    printf("# CURRENT_TIME_UTC=\"%s\"\n",tmpbuf);
  }
	   
  printf("# BGINFO_VERSION=%s\n", BGINFO_VERSION);

  if( bginfo_zepto_version(tmpbuf, sizeof(tmpbuf)) > 0 ) {
    printf("# ZEPTO_VERSION=%s\n", tmpbuf);
  }
  if( level>0 &&
      bginfo_ion_proc_readline("/proc/version", tmpbuf, sizeof(tmpbuf))>0) {
    printf("# ION_LINUX_VERSION=\"%s\"\n", tmpbuf);
  }
  if( bginfo_driver_version(tmpbuf, sizeof(tmpbuf)) > 0 ) {
    printf("# DRIVER_VERSION=%s\n", tmpbuf);
  }
  if( level>0 &&
      bginfo_personality_get_valstr("BGL_VERSION", tmpbuf, sizeof(tmpbuf))>0) {
    printf("# PERSONALITY_VERSION=%s\n", tmpbuf);
  }

  puts("#");

  printf("# N_CNS=%d\n", size);
  printf("# MODE=%s\n", bginfo_cn_mode());
  if( bginfo_is_virtualnodemode() )
    printf("# BGLMPI_MAPPING=%s\n",bginfo_bglmpi_mapping());
  printf("# MEMORY_PER_CN=%d\n",bginfo_memorysize_per_node() );

  if( level>0 ) {
    /* sample usage of BGLPersonality_* function family */
    printf("# TORUS_EXISTS_IN_XYZ=(%d,%d,%d)\n",
	   BGLPersonality_isTorusX(&bginfo.p), 
	   BGLPersonality_isTorusY(&bginfo.p), 
	   BGLPersonality_isTorusZ(&bginfo.p) ); 
    printf("# TORUS_SIZE_XYZ=(%d,%d,%d)\n",
	   BGLPersonality_xSize( &bginfo.p), 
	   BGLPersonality_ySize( &bginfo.p), 
	   BGLPersonality_zSize( &bginfo.p) ); 
    printf("# USE_GLOBAL_INTS=%d\n",
	   BGLPersonality_useGlobalInterrupts( &bginfo.p) ); 
    printf("# PSET_SIZE_XYZ=(%d,%d,%d)\n",
	   BGLPersonality_xPsetSize( &bginfo.p), 
	   BGLPersonality_yPsetSize( &bginfo.p), 
	   BGLPersonality_zPsetSize( &bginfo.p) );
    printf("# N_CNS_IN_PSET=%d\n",
	   BGLPersonality_numNodesInPset( &bginfo.p) );
    printf("# N_PSET=%d\n",
	   BGLPersonality_numPsets( &bginfo.p) );
  }

  puts("#");
  if( bginfo_personality_get_valstr("BGL_BLOCKID", tmpbuf, sizeof(tmpbuf))>0) {
    printf("# BLOCK_ID=%s\n", tmpbuf);
  }
  printf("# N_IONS=%d\n", bginfo_n_ions() );
  if( bginfo_personality_get_valstr("BGL_MTU", tmpbuf, sizeof(tmpbuf))>0) {
    printf("# MTU=%s\n", tmpbuf);
  }
  if( bginfo_personality_get_valstr("BGL_MEMSIZE", tmpbuf, sizeof(tmpbuf))>0) {
    printf("# MEMORY_PER_ION=%s\n", tmpbuf);
  }

  printf("# CN_ION_RATIO=%d\n", bginfo_ci_ratio() );
  printf("# UNEVEN_CI_RATIO=%d\n", bginfo_uneven_ci_ratio() );

  puts("#");


  if( level==0 ) return;

  /* usage of bginfo_struct raw level access */

  bgi = bginfo_get_struct();
  if( ! bgi ) {
    bginfo_perror( "bginfo_get_struct" );
    exit(1);
  }
  puts("#");

  for(i=0;i<bgi->n_ions;i++) {
    printf("# ION_%d_IP=%s\n",i,inet_ntoa(bgi->ions[i].addr));
  }
  puts("#");
  for(i=0;i<bgi->n_ions;i++) {
    printf("# ION_%d_N_CNS=%d\n", i, (bgi->ions[i].ep - bgi->ions[i].sp)+1 );
    printf("# ION_%d_CNS_LIST=\"", i);
    for(j=bgi->ions[i].sp ; j<=bgi->ions[i].ep; j++ ) {
      printf("%d ", bgi->rank_ip_map[j].rank );
    }
    printf("\"\n");
  }
  puts("#");
}

void bginfo_print_info() { _bginfo_print_info(0); }

void bginfo_print_info_full() { _bginfo_print_info(1); }



/*======================================================================
  main function for testing
  ======================================================================*/
#ifdef __TEST_MAIN__
int main(int argc, char* argv[])
{
  int rank;
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  bginfo_print_info_full();
  if( argc >= 2 )     bginfo_print_info();
  if(rank==0) puts("# [done]");
  sleep(60);
  MPI_Finalize();
  return 0;
}
#endif
